"""
Support for OptoDAS format.

This is used by the OptoDAS interrogator made by Alcatel Submarine Networks.

More info here: https://web.asn.com/
"""
from __future__ import annotations
from .core import OptoDASV8, OptoDASV10
