"""
Support for Febus format.

This is used by the Febus DAS interrogator.

More info about febus can be found here: https://www.febus-optics.com/en/
"""
from __future__ import annotations
from .core import Febus2
