"""
Core routines and functionality for processing distributed fiber data.
"""
from __future__ import annotations
from .patch import Patch  # noqa
from .coords import CoordSummary, get_coord
from .coordmanager import get_coord_manager, CoordManager
